/*
 * Copyright (c) 2002, Regents of the University of California. All rights reserved.
 * 
 * ===================================================================
 * The Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        the Institute for Software Research at  
 *        University of California, Irvine"
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and
 *    "University of California" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called
 *    "YANCEES", nor may "YANCEES" appear in their name, without
 *    prior written permission of the University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL WEBCOLLAB.COM OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
 package test;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import edu.uci.isr.yancees.core.MessageParsingException;
import edu.uci.isr.yancees.server.dispatcher.siena.SienaEvent;

public class SienaEventDOMParser {

   // All available DOM Node Types are placed here for convenience
   static final int ELEMENT_TYPE = Node.ELEMENT_NODE;
   static final int ATTR_TYPE = Node.ATTRIBUTE_NODE;
   static final int TEXT_TYPE = Node.TEXT_NODE;
   static final int CDATA_TYPE = Node.CDATA_SECTION_NODE;
   static final int ENTITYREF_TYPE = Node.ENTITY_REFERENCE_NODE;
   static final int ENTITY_TYPE = Node.ENTITY_NODE;
   static final int PROCINSTR_TYPE = Node.PROCESSING_INSTRUCTION_NODE;
   static final int COMMENT_TYPE = Node.COMMENT_NODE;
   static final int DOCUMENT_TYPE = Node.DOCUMENT_NODE;
   static final int DOCTYPE_TYPE = Node.DOCUMENT_TYPE_NODE;
   static final int DOCFRAG_TYPE = Node.DOCUMENT_FRAGMENT_NODE;
   static final int NOTATION_TYPE = Node.NOTATION_NODE;

   public static void main(String argv[]) {

      // Used by the validator parser to use the XML schema grammar.
      final String JAXP_SCHEMA_LANGUAGE =
            "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
      final String W3C_XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";

      // Reference to the DOM tree that will be generated.
      Document document = null;

      if (argv.length != 1) {
         System.err.println("Usage: java EventDOMParser filename");
         System.exit(1);
      }

      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setValidating(true); // validate using grammar specified in document
      factory.setNamespaceAware(true); // also enable namespace separation

      // Other factory options
      factory.setIgnoringComments(true);
      factory.setIgnoringElementContentWhitespace(true);
      factory.setCoalescing(true);
      factory.setExpandEntityReferences(false);

      // Instructs the factory to validate grammar using XML schema.
      try {
         factory.setAttribute(JAXP_SCHEMA_LANGUAGE, W3C_XML_SCHEMA);
      } catch (IllegalArgumentException x) {
         System.out.println("This parser does not support JAXP 1.2");
         System.out.println("Hence, XML schema validation will be turned off");
         System.out.println(x);
      }

      try {
         DocumentBuilder builder = factory.newDocumentBuilder();
         // set error handler before parsing. it is a class to process error messages (optional)
         // instead, we will handle the erros right here, on the spot.
         //builder.setErrorHandler(new MyErrorHandler(System.err));
         document = builder.parse(new File(argv[0]));

      } catch (SAXParseException spe) {
         // Error generated by the parser
         System.out.println("\n** Parsing error"
                            + ", line " + spe.getLineNumber()
                            + ", uri " + spe.getSystemId());
         System.out.println("   " + spe.getMessage());

         // Use the contained exception, if any
         Exception x = spe;
         if (spe.getException() != null) {
            x = spe.getException();
            System.out.println(x);
         }
         //x.printStackTrace();

      } catch (SAXException sxe) {
         // Error generated during parsing)
         Exception x = sxe;
         if (sxe.getException() != null)
            x = sxe.getException();
         System.out.println(sxe);
         //x.printStackTrace();

      } catch (ParserConfigurationException pce) {
         // Parser with specified options can't be built
         System.out.println(pce);
         //pce.printStackTrace();

      } catch (IOException ioe) {
         // I/O error
         System.out.println("EventDOMParser: I/O Error");
         System.out.println(ioe);
         //ioe.printStackTrace();
      }

      System.out.println("Document parsed...");
      printDocument(document);

      // Performs a pre-processing, going to the <event> tag itself
      // and checking if it "looks like" and event file.
      SienaEvent sienaEvt = null;
      if (document != null) {
         NodeList nodes = document.getChildNodes();
         for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE &&
                (node.getNodeName().endsWith(":event") ||
                 node.getNodeName().equals("event"))
                ) {
               try {
                  sienaEvt = new SienaEvent(node);
               } catch (MessageParsingException ex) {
                  System.out.println(ex);
               }
               break;
            }
         }
      }

      if (sienaEvt != null) {
         System.out.println("Getting siena event...");
         System.out.println(sienaEvt);

      } else {
         System.out.println("");
         System.out.println(
               "Could not parse the document and get <event> tag.");
      }

   } // main



   /**
    * Recursively parsers the dom model, using the facilities of the
    * AdapterNode class, and prints it to the standard output.
    */
   private static void printDocument(org.w3c.dom.Node n) {
      System.out.println("");
      printDocumentAux("", n);
      System.out.println("");

   }

   /**
    * This method provides ident formating to the print procedure.
    */
   private static void printDocumentAux(String ident, org.w3c.dom.Node n) {
      String identStep = "   ";
      if (n != null) {
         int type = n.getNodeType();

         switch (type) {
            case ELEMENT_TYPE:
               System.out.print(ident + "<" + n.getNodeName());
               if (n.hasAttributes()) {
                  System.out.print(" ");
                  NamedNodeMap map = n.getAttributes();
                  for (int i = 0; i < map.getLength(); i++) {
                     printDocumentAux(ident, map.item(i));
                  }
               }
               System.out.println(">");

               if (n.hasChildNodes()) {
                  NodeList list = n.getChildNodes();
                  for (int i = 0; i < list.getLength(); i++) {
                     printDocumentAux(ident + identStep, list.item(i));
                  }
               }
               System.out.print(ident + "</" + n.getNodeName() + ">");
               break;
            case ATTR_TYPE:
               System.out.print(n.getNodeName() + " = \"" + n.getNodeValue() +
                                "\" ");
               break;
            case TEXT_TYPE:
               System.out.println(ident + n.getNodeValue().trim());
               break;

            default:

               // Recursively call the child.
               if (n.hasChildNodes()) {
                  NodeList list = n.getChildNodes();
                  for (int i = 0; i < list.getLength(); i++) {
                     printDocumentAux(ident, list.item(i));
                  }
               }
         }

      } else {
         System.out.println("ERROR: document was not parsed properly...");
      }

   }





   /**
    * Error handler to report errors and warnings.
    * Allows the customization of the parsing error messages.
    */
   private static class MyErrorHandler
         implements ErrorHandler {
      /** Error handler output goes here */
      private PrintStream out;

      MyErrorHandler(PrintStream out) {
         this.out = out;
      }

      /**
       * Returns a string describing parse exception details
       * The DOM parser uses the SAX parse exception to convey its errors
       */
      private String getParseExceptionInfo(SAXParseException spe) {
         String systemId = spe.getSystemId();
         if (systemId == null) {
            systemId = "null";
         }
         String info = "URI=" + systemId +
               " Line=" + spe.getLineNumber() +
               ": " + spe.getMessage();
         return info;
      }

      // The following methods are standard SAX ErrorHandler methods.
      // See SAX documentation for more info.

      public void warning(SAXParseException spe) throws SAXException {
         out.println("Warning: " + getParseExceptionInfo(spe));
      }

      public void error(SAXParseException spe) throws SAXException {
         String message = "Error: " + getParseExceptionInfo(spe);
         throw new SAXException(message);
      }

      public void fatalError(SAXParseException spe) throws SAXException {
         String message = "Fatal Error: " + getParseExceptionInfo(spe);
         throw new SAXException(message);
      }
   }

}