/*
 * Copyright (c) 2002, Regents of the University of California. All rights reserved.
 * 
 * ===================================================================
 * The Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        the Institute for Software Research at  
 *        University of California, Irvine"
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and
 *    "University of California" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called
 *    "YANCEES", nor may "YANCEES" appear in their name, without
 *    prior written permission of the University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL WEBCOLLAB.COM OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
 /*
 * Created on Sep 1, 2003
 *
 * <p>Title: Yancees Notification Server</p>
 * <p>Description: Yet ANother Configurable Extensible Event Service</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: School of Information and Computer Science - University of California, Irvine</p>
 * @author Roberto Silveira Silva Filho
 * @version 1.0
 */
package edu.uci.isr.yancees.filter;

import edu.uci.isr.yancees.EventInterface;


/**
 * Implements an abstract filter with the basic functionality to the 
 * methods of the FilterInterface
 */
public abstract class AbstractFilter implements FilterInterface {

	private FilterInterface successor;
	private String filterName;
	
	private boolean print = edu.uci.isr.yancees.YanceesProperties.getInstance().PRINT_DEBUG;
	
	/**
	 * constructor
	 */
	public AbstractFilter() {
		super();
		filterName = null;
		successor = null;
	}

	/**
	 * Handles the incomming message and passes it forward to all the 
	 * successors of this filter
	 * @param evt is the event to process
	 * @see yancees.filter.FilterInterface#handleMessage(yancees.core.MessageInterface)
	 */
	public void handleMessage(EventInterface evt) {
		if (print)
			System.out.println("AbstractFilter: Handling event at filter: "+this.getName());	
		
		if (evt != null) {
			EventInterface[] filteredEvents = doFilterEvent(evt);
			if (successor != null) {
				successor.handleMessage(filteredEvents);
			}
		}
	}
	
	/**
	 * Handles the incomming message and passes it forward to all the 
	 * successors of this filter
	 * @param evtList is the list of events to process
	 * @see yancees.filter.FilterInterface#handleMessage(yancees.core.MessageInterface)
	 */
	public void handleMessage(EventInterface[] evtList) {
		
		EventInterface[] responseList;
		
		if (print)
			System.out.println("AbstractFilter: Handling event list at filter: "+this.getName());
			
		if (evtList != null || evtList.length > 0) {
			responseList = doFilterEventList(evtList);
			if (successor != null) {
				successor.handleMessage(responseList);
			}
		}
	}

	/**
	 * This method should be implemented by the subclasses of the AbstractFilter
	 * in order to process the messages being passed through this filter.
	 * A filter uses this method to receive the message and process it. The filter
	 * must then select which events to let go to the next filter or subscriber. 
	 * These events are returned in the method.
	 * 
	 * @param evt is the event to be filtered
	 * @return is the list of events resulted from the filtering
	 */
	protected abstract EventInterface[] doFilterEvent(EventInterface evt);

	/**
	 * This method should be implemented by the subclasses of the AbstractFilter
	 * in order to process the messages being passed through this filter.
	 * 
	 * A filter uses this method to receive the message and process it. The filter
	 * must then select which events to let go to the next filter or subscriber. 
	 * These events are returned in the method.
	 
	 * @param evtList is a list of events to be filtered as a "burst"
	 * @return is the list of events resulted from the filtering
	 */
	protected abstract EventInterface[] doFilterEventList(EventInterface[] evtList);


	/**
    * Adds a successor to the list of filters of this filter
	 * @see yancees.filter.FilterInterface#addSucessor(yancees.filter.FilterInterface)
	 */
	public void addSuccessor(FilterInterface filter) {
		successor = filter;
	}

	/** 
    * Remove a filter from the list of successors from this filter
	 * @see yancees.filter.FilterInterface#removeSucessor(yancees.filter.FilterInterface)
	 */
	public void removeSuccessor() {
		successor = null;
	}

	/**
	 * list all the successors of this filters if any
	 * @see yancees.filter.FilterInterface#getSucessorsList()
	 * @return a list of filgers or an empty list if no successors are registered
	 */
	public FilterInterface getSuccessor() {
		return successor;
	}

	/**
	 * @see yancees.filter.FilterInterface#getName()
	 * @return the name of this filter
	 */
	public String getName() {
		return filterName;
	}
	
	public void setName(String name) {
		filterName = name;
	}
	
}
